@@ -9,7 +9,8 @@
 #
 # Released into public domain.
 
-my $version = "2.37";
+my $version = "2.40";
+my $maintainer = "kjetilho+listadmin\@ifi.uio.no";
 
 use HTML::TokeParser;
 use LWP::UserAgent;
@@ -187,7 +188,7 @@ for (@lists) {
     if ($info->{'servererror'}) {
 	print "\n";
 	printf STDERR ("ERROR: fetching %s\n", $info->{'url'});
-	printf STDERR ("ERROR: Server returned '%s' -- skipping list\n",
+	printf STDERR ("ERROR: %s -- skipping list\n",
 		       $info->{'servererror'});
 	next;
     } elsif ($info->{'autherror'}) {
@@ -252,7 +253,7 @@ sub process_subscriptions {
 	}
     }
     my $count = keys (%subscribers);
-    my $def = $config->{"subdef"};
+    my $def = $config->{"subdefault"};
     my $prompt = 'Accept/Reject/Skip/Quit';
     $prompt .= " [" . uc($def) . "]" if $def;
     $prompt .= " ? ";
@@ -266,7 +267,7 @@ sub process_subscriptions {
 	print "         subscription request\n";
 	my $ans;
 	while (1) {
-	    $ans = $config->{"subact"};
+	    $ans = $config->{"subaction"};
 	    $ans ||= prompt ($prompt);
 	    $ans = "q" unless defined $ans;
 	    $ans =~ s/\s+//g;
@@ -647,29 +648,48 @@ sub get_list {
 	return {'servererror' => $resp->status_line, 'url' => $url};
     }
     $page = $resp->content;
-    if ($page eq "") {
-	if (time - $starttime > 60) {
-	    print "Mailman server timed out?\n";
-	} else {
-	    print "empty page\n";
-	}
-	return {};
-    }
+
     my $dumpdir = $config->{"dumpdir"};
-    if (defined $dumpdir) {
-	if (open (DUMP, ">$dumpdir/dump-$list.html")) {
+    my $dumpfile;
+    if ($dumpdir && $page) {
+	$dumpfile = "$dumpdir/dump-$list.html";
+	if (open (DUMP, ">$dumpfile")) {
 	    print DUMP $page;
 	    close (DUMP);
 	}
     }
 
+    if ($page eq "") {
+	if (time - $starttime > 60) {
+	    return {servererror => "Mailman server timed out?", url => $url};
+	} else {
+	    return {servererror => "Empty page", url => $url};
+	}
+    } elsif ($page =~ get_trans_re("no_such_list")) {
+	return {servererror => "No such list", url => $url}
+    }
+
     my $parse = HTML::TokeParser->new(\$page) || die;
     $parse->get_tag ("title") || die;
     my $title = $parse->get_trimmed_text ("/title") || die;
-    if ($title =~ /authentication/i) {
+
+    if ($title =~ get_trans_re("authentication")) {
 	return {'autherror' => 1};
     }
 
+    if ($page !~ get_trans_re("pending_req")) {
+	my $msg = "unexpected contents";
+	# Use rand() to protect a little against tmpfile races
+	$dumpfile ||= "/tmp/dump-" . rand() . "-$list.html";
+	if (open(DUMP, ">$dumpfile")) {
+	    chmod(0600, $dumpfile);
+	    print DUMP $page;
+	    close(DUMP);
+	    $msg .= ", please send $dumpfile to $maintainer";
+	}
+	return {servererror => $msg, url => $url};
+    }
+
     my @mailman_mentions = grep {/Mailman/} split (/\n/, $page);
     for my $mention (reverse @mailman_mentions) {
 	if ($mention =~ /\bv(ersion)?\s(\d+\.\d+)/) {
@@ -678,10 +698,7 @@ sub get_list {
 	}
     }
     unless ($mmver) {
-	if ($page =~ /no such list/i) {
-	    return {'servererror' => "No such list", 'url' => $url}
-	}
-	die "Can not find version information in, please mail maintainer.";
+	die "Can not find version information, please mail maintainer.";
     }
 
     if ($mmver ge "2.1") {
@@ -719,7 +736,7 @@ sub parse_pages_mm_old {
     $parse->get_tag ("hr");
     $parse->get_tag ("h2") || return \%data;
     my $headline = $parse->get_trimmed_text ("/h2") || die;
-    if ($headline =~ /subscription/i) {
+    if ($headline =~ get_trans_re("headline_subscr")) {
 	parse_subscriptions ($mmver, $config, $parse, \%data);
 	$token = $parse->get_token;
 	if (lc ($token->[1]) eq "input") {
@@ -729,7 +746,7 @@ sub parse_pages_mm_old {
 	    $headline = $parse->get_trimmed_text ("/h2") || die;
 	}
     }
-    if ($headline =~ /held for approval/i) {
+    if ($headline =~ get_trans_re("held_for_approval")) {
 	parse_approvals ($mmver, $config, $parse, \%data);
     } else {
 	$parse->get_tag ("hr") || die;
@@ -801,6 +818,62 @@ sub parse_approvals {
     } until ($token->[0] eq "S" && lc ($token->[1]) eq "input");
 }
 
+sub get_trans_re {
+    my ($key) = @_;
+
+    # Handle translations -- poorly...
+    #
+    # For now, we look for strings in all languages at the same time
+    # since they don't seem to overlap.  This might have to change
+    # later.
+    #
+    # Please send additions if you have them.
+
+    my %translations =
+	    ("authentication" =>
+	     {
+	      "en" => "authentication",
+	      "de" => "Authentifikation",
+	      "fr" => "authentification",
+	     },
+	     "subscr_success" =>
+	     {
+	      "en" => "Successfully ((un)?subscribed|Removed)",
+	      "de" => "Erfolgreich (ein|aus)getragen",
+	     },
+	     "subscr_error" =>
+	     {
+	      "en" => "Error (un)?subscribing",
+	     },
+	     "no_such_list" =>
+	     {
+	      "en" => "Mailman Admindb Error.*No such list:",
+	     },
+	     "pending_req" =>
+	     {
+	      "en" => "(current set of administrative|pending request)",
+	      "de" => "(gegenw&auml;rtigen administrativen|unbearbeiteten Anfragen)",
+	     },
+	     "headline_subscr" =>
+	     {
+	      "en" => "subscription",
+	     },
+	     "held_for_approval" =>
+	     {
+	      "en" => "held for approval",
+	     },
+	     "already_member" =>
+	     {
+	      "en" => "Already a member",
+	     },
+	    );
+
+    my $t = $translations{$key};
+    die "INTERNAL ERROR: Unknown translation key '$key'\n"
+	    unless defined $t;
+    return "(?i)(" . join("|", values %{$t}) . ")";
+}
+
 sub guess_charset {
     my ($charset, $text) = @_;
 
@@ -1023,29 +1096,24 @@ sub set_param_values {
 sub read_config {
     my ($file) = @_;
 
-    my ($user, $pw, $spam, $list);
+    my %cur = map { $_ => undef; }
+	    qw (not_spam_if_from
+		not_spam_if_subject
+		discard_if_from
+		discard_if_subject
+		discard_if_reason);
+    my $pattern_keywords = join ("|", keys %cur);
+
+    # Defaults:
+    $cur{user} = $cur{password} = $cur{action} = $cur{default} = "";
+    $cur{confirm} = 1;
+    $cur{unprintable} = "questionmark";
+
     my $conf = {};
     my $line = "";
-    my $subact;
-    my $subdef;
-    my $action = "";
-    my $default = "";
     my $count = 0;
     my $lineno = 0;
-    my $logfile;
-    my $dumpdir;
-    my $confirm = 1;
-    my $url;
-    my $spamheader;
-    my $unprintable = "questionmark";
-    my %patterns = map { $_ => undef; }
-                       qw (not_spam_if_from
-			   not_spam_if_subject
-			   discard_if_from
-			   discard_if_subject
-			   discard_if_reason);
-    my $pattern_keywords = join ("|", keys %patterns);
-    
+
     my %act = ("approve" => "a", "discard" => "d",
 	       "reject" => "r", "skip" => "s", "none" => "");
     my %sact = ("accept" => "a",
@@ -1067,77 +1135,61 @@ sub read_config {
 	$line =~ s/^\s+//;
 	next if /^$/;
 	if ($line =~ /^username\s+/i) {
-	    $user = unquote($POSTMATCH);
-	    if ($user !~ /^[a-z0-9._+-]+\@[a-z0-9.-]+$/) {
-		print STDERR "$file:$lineno: Illegal username: '$user'\n";
+	    $cur{user} = unquote($POSTMATCH);
+	    if ($cur{user} !~ /^[a-z0-9._+-]+\@[a-z0-9.-]+$/) {
+		print STDERR "$file:$lineno: Illegal username: '$cur{user}'\n";
 		exit 1;
 	    }
 	} elsif ($line =~ /^password\s+/i) {
-	    $pw = unquote($POSTMATCH);
+	    $cur{password} = unquote($POSTMATCH);
 	} elsif ($line =~ /^spamlevel\s+/i) {
-	    $spam = unquote($POSTMATCH);
-	    if ($spam =~ /^(\d+)\s*$/) {
-		$spam = $1;
+	    $cur{spamlevel} = unquote($POSTMATCH);
+	    if ($cur{spamlevel} =~ /^(\d+)\s*$/) {
+		$cur{spamlevel} = $1;
 	    } else {
-		print STDERR "$file:$lineno: Illegal value: '$spam'\n";
+		print STDERR "$file:$lineno: Illegal value: '$cur{spamlevel}'\n";
 		print STDERR "choose a positive numeric value\n";
 		exit 1;
 	    }
-	} elsif ($line =~ /^confirm\s+/i) {
-	    $confirm = unquote($POSTMATCH);
-	    if ($confirm eq "yes") {
-		$confirm = 1;
-	    } elsif ($confirm eq "no") {
-		$confirm = undef;
+	} elsif ($line =~ /^(confirm|meta_member_support)\s+/i) {
+	    my ($key, $value) = (lc($1), unquote($POSTMATCH));
+	    if ($value eq "yes") {
+		$value = 1;
+	    } elsif ($value eq "no") {
+		$value = undef;
 	    } else {
-		print STDERR "$file:$lineno: Illegal value: '$confirm'\n";
+		print STDERR "$file:$lineno: Illegal value: '$value\n";
 		print STDERR "choose one of yes or no\n";
 		exit 1;
 	    }
-	} elsif ($line =~ /^action\s+/i) {
-	    $action = unquote($POSTMATCH);
-	    unless (exists $act{$action}) {
-		print STDERR "$file:$lineno: Illegal value: '$action'\n";
+	    $cur{$key} = $value;
+	} elsif ($line =~ /^(action|default)\s+/i) {
+	    my ($key, $value) = (lc($1), unquote($POSTMATCH));
+	    unless (exists $act{$value}) {
+		print STDERR "$file:$lineno: Illegal value: '$value\n";
 		print STDERR "choose one of ",
                              join (", ", sort keys %act), "\n";
 		exit 1;
 	    }
-	    $action = $act{$action};
+	    $cur{$key} = $act{$value};
 	} elsif ($line =~ /^adminurl\s+/i) {
-	    $url = unquote($POSTMATCH);
-	    $url = undef if $url eq "NONE";
-	} elsif ($line =~ /^default\s+/i) {
-	    $default = unquote($POSTMATCH);
-	    unless (exists $act{$default}) {
-		print STDERR "$file:$lineno: Illegal value: '$default'\n";
-		print STDERR "choose one of ",
-                             join (", ", sort keys %act), "\n";
-		exit 1;
-	    }
-	    $default = $act{$default};
+	    $cur{adminurl} = unquote($POSTMATCH);
+	    $cur{adminurl} = undef if $cur{adminurl} eq "NONE";
 	} elsif ($line =~ /^log\s+/i) {
-	    $logfile = expand_pathname(unquote($POSTMATCH));
+	    $cur{logfile} = expand_pathname(unquote($POSTMATCH));
 	} elsif ($line =~ /^dumpdir\s+/i) {
-	    $dumpdir = expand_pathname(unquote($POSTMATCH));
-	    mkdir($dumpdir) if (defined $dumpdir);
-	} elsif ($line =~ /^subscription_action\s+/) {
-	    $subact = unquote($POSTMATCH);
-	    unless (exists $sact{$subact}) {
-		print STDERR "$file:$lineno: Illegal value: '$subact'\n";
+	    $cur{dumpdir} = expand_pathname(unquote($POSTMATCH));
+	    mkdir($cur{dumpdir}) if (defined $cur{dumpdir});
+	} elsif ($line =~ /^subscription_(action|default)\s+/) {
+	    my $key = "sub" . lc($1);
+	    my $value = unquote($POSTMATCH);
+	    unless (exists $sact{$value}) {
+		print STDERR "$file:$lineno: Illegal value: '$value'\n";
 		print STDERR "choose one of ",
 		             join (", ", sort keys %sact), "\n";
 		exit 1;
 	    }
-	    $subact = $sact{$subact};
-	} elsif ($line =~ /^subscription_default\s+/) {
-	    $subdef = unquote($POSTMATCH);
-	    unless (exists $sact{$subdef}) {
-		print STDERR "$file:$lineno: Illegal value: '$subdef'\n";
-		print STDERR "choose one of ",
-		             join (", ", sort keys %sact), "\n";
-		exit 1;
-	    }
-	    $subdef = $sact{$subdef};
+	    $cur{$key} = $sact{$value};
 	} elsif ($line =~ /^($pattern_keywords)\s+/o) {
 	    my $key = $1;
 	    my $val = $POSTMATCH;
@@ -1147,37 +1199,24 @@ sub read_config {
 		$val =~ s/\\"/"/g;
 		$val =~ s/\\\\/\\/g;
 	    }
-	    $patterns{$key} = ($val eq "NONE") ? undef : $val;
+	    $cur{$key} = ($val eq "NONE") ? undef : $val;
 	} elsif ($line =~ /^spamheader\s+/) {
-	    $spamheader = unquote($POSTMATCH);
-	    unless ($spamheader =~ /^[\w-]+$/) {
+	    $cur{spamheader} = unquote($POSTMATCH);
+	    unless ($cur{spamheader} =~ /^[\w-]+$/) {
 		print STDERR "$file:$lineno: Illegal header name: ".
-			"'$spamheader'\n";
+			"'$cur{spamheader}'\n";
 		exit 1;
 	    }
-	    $spamheader = undef if $spamheader eq "default";
+	    $cur{spamheader} = undef if $cur{spamheader} eq "default";
 	} elsif ($line =~ /^([^@ \t]+@[^@])+\s*/) {
-	    $conf->{$line} = { "user" => $user,
-			       "password" => $pw,
-			       "adminurl" => $url,
-			       "spamlevel" => $spam,
-			       "confirm" => $confirm,
-			       "subact" => $subact,
-			       "subdef" => $subdef,
-			       "action" => $action,
-			       "default" => $default,
-			       "logfile" => $logfile,
-			       "dumpdir" => $dumpdir,
-			       "spamheader" => $spamheader,
-			       "unprintable" => $unprintable,
-			       %patterns,
-			       "order" => ++$count,
-			   };
+	    my %copy = %cur;
+	    $copy{order} = ++$count;
+	    $conf->{$line} = \%copy;
 	} elsif ($line =~ /^unprintable\s+/) {
-	    $unprintable = unquote($POSTMATCH);
-	    unless ($unprintable =~ /^(questionmark|unicode)$/) {
+	    $cur{unprintable} = unquote($POSTMATCH);
+	    unless ($cur{unprintable} =~ /^(questionmark|unicode)$/) {
 		print STDERR "$file:$lineno: Illegal format for ".
-			"unprintable characters: '$unprintable'\n";
+			"unprintable characters: '$cur{unprintable}'\n";
 		exit 1;
 	    }
 	} else {
@@ -1301,6 +1340,8 @@ sub commit_changes {
     my $params = mailman_params ($user, $pw);
 
     my $log = log_timestamp ($list);
+    # Expand {list}, {subdomain} and {domain}
+    $logfile = mailman_url($list, $logfile);
 
     for my $id (sort { $a <=> $b } keys %{$change}) {
 	my ($what, $text) = @{$change->{$id}};
@@ -1359,12 +1400,15 @@ sub add_subscribers {
 
     die unless @addresses;
 
+    fetch_meta_members($list, $config);
+
     my %params = (username => $config->{user},
 		  adminpw => $config->{password},
 		  subscribe_or_invite => 0,                # Mailman 2.x
 		  send_notifications_to_list_owner => 0,   # Mailman 2.x
 		  send_welcome_message_to_this_batch => 0, # Mailman 2.x
 		  send_welcome_msg_to_this_batch => 0,     # Mailman 1.2
+		  meta_members => $config->{meta_members}, # Mailman 1.2
 		  subscribees => join("\n", @addresses));
     my $url = mailman_url($list, $config->{adminurl}, "", "members");
     my $resp = $ua->post($url, \%params);
@@ -1375,7 +1419,7 @@ sub add_subscribers {
     if (!$mail) {
 	my %left = map { $_ => 1 } @addresses;
 	for my $failed (keys %{$result}) {
-	    unless ($result->{$failed} =~ /Already a member/) {
+	    unless ($result->{$failed} =~ get_trans_re("already_member")) {
 		delete $left{$failed};
 	    }
 	}
@@ -1385,7 +1429,7 @@ sub add_subscribers {
 	# members.
 	@addresses = ();
 	for my $failed (keys %{$result}) {
-	    if ($result->{$failed} =~ /Already a member/) {
+	    if ($result->{$failed} =~ get_trans_re("already_member")) {
 		push(@addresses, $failed);
 	    }
 	}
@@ -1394,6 +1438,7 @@ sub add_subscribers {
 	%params = (username => $config->{user},
 		   adminpw => $config->{password},
 		   user => \@addresses,
+		   meta_members => $config->{meta_members}, # Mailman 1.2
 		   setmemberopts_btn => "submit");	# Mailman 2.x
 	for my $a (@addresses) {
 	    $params{$a . "_nomail"} = "on" unless $mail;
@@ -1409,6 +1454,8 @@ sub add_subscribers {
 sub remove_subscribers {
     my ($list, $config, @addresses) = @_;
 
+    fetch_meta_members($list, $config);
+
     my $url = mailman_url($list, $config->{adminurl}, "", "members");
 
     # In Mailman 1.2, unsubscription happens when an address is
@@ -1417,6 +1464,7 @@ sub remove_subscribers {
     my %params = (username => $config->{user},
 		  adminpw => $config->{password},
 		  setmemberopts_btn => "submit",	# Mailman 2.x
+		  meta_members => $config->{meta_members}, # Mailman 1.2
 		  user => \@addresses);
     for my $a (@addresses) {
 	$params{$a . "_unsub"} = "on";			# Mailman 2.x
@@ -1449,9 +1497,9 @@ sub parse_subscribe_response {
 	$parse->get_tag ("ul") || die;
 	my $ul = $parse->get_text ("/ul") || die;
 
-	if ($h5 =~ /Successfully ((un)?subscribed|Removed)/i) {
+	if ($h5 =~ get_trans_re("subscr_success")) {
 	    # hooray!
-	} elsif ($h5 =~ /Error (un)?subscribing/i) {
+	} elsif ($h5 =~ get_trans_re("subscr_error")) {
 	    for (split(/\n/, $ul)) {
 		chomp;
 		if (/^\s*(.*?)\s*--\s*(.*)/) {
@@ -1461,7 +1509,7 @@ sub parse_subscribe_response {
 	} else {
 	    $ul =~ s/\n/\n\t/g;
 	    print STDERR "You have an unusual Mailman output.  Please mail ".
-		    "this message to\nkjetilho+listadmin\@ifi.uio.no:\n".
+		    "this message to\n$maintainer\n:\n".
 		    "\t[$h5]\n\t[$ul]\nThanks!\n";
 	}
 	$parse->get_tag ("p") || die;
@@ -1473,9 +1521,11 @@ sub parse_subscribe_response {
 sub list_subscribers {
     my ($list, $config) = @_;
 
+    fetch_meta_members($list, $config);
     my $url = mailman_url($list, $config->{adminurl}, "", "members");
     my %params = (username => $config->{user},
 		  adminpw => $config->{password},
+		  meta_members => $config->{meta_members},
 		  chunk => 0);
     my $resp = $ua->post($url, \%params);
     unless ($resp->is_success) {
@@ -1549,9 +1599,52 @@ sub list_subscribers {
 	    $resp = $ua->post("$url?letter=$letter&chunk=$chunk", \%params);
 	}
     }
+    if ($config->{meta_members}) {
+	push(@addresses, split(/\n+/, $config->{meta_members}));
+    }
     return @addresses;
 }
 
+# This code is only useful on the patched Mailman 1.2 installation at
+# UiO.  Notice that it uses GET without any parameters to fetch the
+# page, since otherwise it will clear the meta members.
+# Unfortunately, this means we need to use cookies to log in, and this
+# requires a new Perl module, WWW::Mechanize.  Since this is such a
+# site specific feature, we hide the requirement so listadmin runs
+# even without the module.
+
+sub fetch_meta_members {
+    my ($list, $config) = @_;
+
+    return if defined $config->{meta_members}; # already fetched
+    return unless $config->{meta_member_support} || $list =~ /\buio\.no$/i;
+
+    # We will only attempt this once, so make a note we've tried.
+    $config->{meta_members} = "";
+
+    unless (eval "require WWW::Mechanize; 1") {
+	print "WARNING: Meta members may be removed, install WWW::Mechanize\n";
+	return;
+    }
+
+    my $agent = WWW::Mechanize->new(autocheck => 1);
+    $agent->get(mailman_url($list, $config->{adminurl}));
+    $agent->submit_form(fields => { username => $config->{user},
+				    adminpw => $config->{password}});
+
+    $agent->get(mailman_url($list, $config->{adminurl}, "", "members"));
+
+    my $page = $agent->content();
+    my $parse = HTML::TokeParser->new(\$page);
+    my $tag = $parse->get_tag("textarea");
+    $tag = $parse->get_tag("textarea");
+    return unless defined $tag; # silently ignore the failure
+
+    if ($tag->[1]->{name} eq "meta_members") {
+	$config->{meta_members} = $parse->get_trimmed_text("/textarea");
+    }
+}
+
 sub remove_matching_subscribers {
     my ($list, $config, $pattern) = @_;
     my @addresses = list_subscribers($list, $config);
@@ -1589,7 +1682,9 @@ sub submit_http {
     if ($logfile) {
 	if (open (LOG, ">>$logfile")) {
 	    LOG->autoflush(1);
-	    binmode LOG, ":encoding(" . langinfo(CODESET()) . ")";
+	    # Perhaps we should force the encoding to US-ASCII
+	    # instead, but I think this is more DWIM compliant.
+	    binmode LOG, ":encoding($term_encoding)";
 	    $opened = 1;
 	    local $SIG{__WARN__} = sub {}; # see comment elsewhere
 	    print LOG $log;
@@ -1634,6 +1729,9 @@ sub prompt_password {
     my $answer;
     my $echooff;
 
+    # This might not work, since some versions of readline screw up
+    # and turn on "echo" for us :-(
+
     $SIG{'INT'} = $SIG{'TERM'} = \&restore_echo_and_exit;
     system("stty -echo 2>/dev/null");
     if ($? == 0) {
@@ -1655,7 +1753,10 @@ sub prompt {
     # is only done if the user actually needs prompting.
     $term = new Term::ReadLine 'listadmin'
 	    unless $term;
-    return ($term->readline (@_));
+    my $answer = $term->readline(@_);
+    # readline turns off autoflush, re-enable it
+    $| = 1;
+    return $answer;
 }
 
 sub config_order {
